/*
PeRdr - PE file disassembler
Copyright (C) 1999-2002 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#ifndef FILE_BYTELEVEL_H
#define FILE_BYTELEVEL_H

#include "rva.h"
#include "file.hpp"
#include "utils/range.hpp"
#include "relocationinfo.h"

// gestisce il file come se fosse caricato in memoria
// throw exception
class FileByte
{
public:
  class Section: public range<vma_t>
  {
  public:
    Section(): range<vma_t>(0,0),flag(0) {};
    // they was bool, but this cause some warning
    unsigned IsCode() const { return flag & fCode; }
    unsigned IsData() const { return flag & fData; }
    void SetIsCode() { flag |= fCode; }
    void SetIsData() { flag |= fData; }
  private:
    enum { fCode = 1, fData = 2, fWritable = 4 };
    unsigned flag;
  };
  typedef std::vector<Section> TSections;

  FileByte(CFile& file,const RVAFileTranslator& rva,uint32_t imageBase,bool hasRelocation,const RelocationInfos& relocationInfos);
  ~FileByte();
  struct OutOfAddress
  {
    OutOfAddress(vma_t addr):address(addr) {};
    vma_t address;
  };
  uint8_t ReadByte();
  uint8_t ReadByte(vma_t address);
  uint16_t ReadWord();
  uint16_t ReadWord(vma_t address);
  uint32_t ReadDword();
  uint32_t ReadDword(vma_t address);
  void SeekTo(vma_t address);
  vma_t Tell();
  bool IsValid(vma_t address);
  vma_t GetNextValid();

  // aggiunti per "compatibilita'" bfd
  bool HasRelocation() const
  { return hasRelocation; }
  const RelocationInfos& GetRelocationInfos() const
  { return relocationInfos; }

  // section access
  const Section* GetCurrentSection() const { return &*currSection; }
  // se non trovata nessuna sezione ritorna GetErrorSection()
  const Section* GetSection(vma_t address) const;
  // sezione usata per ritorno in caso di sezione non trovata
  static const Section* GetErrorSection() { return &errorSection; }

  //
  const TSections& GetSections() const
  { return sections; }
private:
//  int unsafeReadByte(vma_t address);
  uint8_t **sectionBuffer;
  uint8_t *pSectionBuffer;
//  typedef std::vector<Section> TSections;
  TSections sections;
  Section* currSection;
  vma_t currPos;
  // not implemented
  void operator=(const FileByte&);
  FileByte(const FileByte&);
  static const Section errorSection;

  bool hasRelocation;
  RelocationInfos relocationInfos;
};
#endif // FILE_BYTELEVEL_H
